<?php

namespace Drupal\search_api_workspaces\Plugin\search_api\datasource;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\search_api\Plugin\search_api\datasource\ContentEntityTrackingManager;
use Drupal\search_api\SearchApiException;
use Drupal\search_api\Utility\Utility;

class WorkspacesContentEntityTrackingManager extends ContentEntityTrackingManager {

  /**
   * {@inheritDoc}
   */
  public function entityInsert(EntityInterface $entity) {
    // Check if the entity is a content entity.
    if (!($entity instanceof ContentEntityInterface) || $entity->search_api_skip_tracking) {
      return;
    }
    $indexes = $this->getIndexesForEntity($entity);
    if (!$indexes) {
      return;
    }

    // Compute the item IDs for all languages of the entity.
    $item_ids = [];
    $entity_id = $entity->id();
    $revision_id = $entity->get('vid')->value;
    foreach (array_keys($entity->getTranslationLanguages()) as $langcode) {
      $item_ids[] = $entity_id . ':' . $langcode . ':' . $revision_id;
    }
    $datasource_id = 'workspaces_entity:' . $entity->getEntityTypeId();
    foreach ($indexes as $index) {
      $filtered_item_ids = $this->filterValidItemIds($index, $datasource_id, $item_ids);
      $index->trackItemsInserted($datasource_id, $filtered_item_ids);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function entityUpdate(EntityInterface $entity) {
    // Check if the entity is a content entity.
    if (!($entity instanceof ContentEntityInterface) || $entity->search_api_skip_tracking) {
      return;
    }
    $indexes = $this->getIndexesForEntity($entity);

    if (!$indexes) {
      return;
    }

    // Compare old and new languages for the entity to identify inserted,
    // updated and deleted translations (and, therefore, search items).
    // We enforce creation of revisions always delete revision so no update.
    $entity_id = $entity->id();
    $old_revision_id = $entity->original->getRevisionId();

    // Remove info about the merged entity revision during sync.
    if ($entity->isSyncing() && !$entity->original->revision_merge_parent->isEmpty()) {
      $old_revision_id = $entity->original->revision_merge_parent->target_revision_id;
    }

    $revision_id = $entity->getRevisionId();
    // Build item ids based on translations.
    $insert_item_ids = array_keys($entity->getTranslationLanguages());
    $deleted_item_ids = [];
    $old_translations = $entity->original->getTranslationLanguages();

    foreach ($old_translations as $langcode => $language) {
      $deleted_item_ids[] = $langcode;
    }

    // Soft delete.
    if ($entity->hasField('deleted') && $entity->get('deleted')->value) {
      foreach ($insert_item_ids as $key => $langcode) {
        // Don't add this as a newly indexed reivision as it's soft-deleted.
        unset($insert_item_ids[$key]);
      }
    }

    $datasource_id = 'workspaces_entity:' . $entity->getEntityTypeId();

    $combine_id_old = function ($langcode) use ($entity_id, $old_revision_id) {
      return $entity_id . ':' . $langcode . ':' . $old_revision_id;
    };
    $combine_id_new = function ($langcode) use ($entity_id, $revision_id) {
      return $entity_id . ':' . $langcode . ':' . $revision_id;
    };

    $insert_item_ids = array_map($combine_id_new, $insert_item_ids);
    $deleted_item_ids = array_map($combine_id_old, $deleted_item_ids);

    foreach ($indexes as $index) {
      if ($insert_item_ids) {
        $filtered_item_ids = $this->filterValidItemIds($index, $datasource_id, $insert_item_ids);
        $index->trackItemsInserted($datasource_id, $filtered_item_ids);
      }
      if ($deleted_item_ids) {
        $index->trackItemsDeleted($datasource_id, $deleted_item_ids);
      }
    }
  }

  /**
   * {@inheritDoc}
   */
  public function entityDelete(EntityInterface $entity) {
    // Check if the entity is a content entity.
    if (!($entity instanceof ContentEntityInterface) || $entity->search_api_skip_tracking) {
      return;
    }
    $indexes = $this->getIndexesForEntity($entity);
    if (!$indexes) {
      return;
    }

    // Remove the search items for all the entity's translations.
    $item_ids = [];
    $entity_id = $entity->id();
    $revision_id = $entity->getRevisionId();
    foreach (array_keys($entity->getTranslationLanguages()) as $langcode) {
      $item_ids[] = $entity_id . ':' . $langcode . ':' . $revision_id;
    }
    $datasource_id = 'workspaces_entity:' . $entity->getEntityTypeId();

    foreach ($indexes as $index) {
      $index->trackItemsDeleted($datasource_id, $item_ids);
    }
  }

  /**
   * {@inheritDoc}
   */
  public function getIndexesForEntity(ContentEntityInterface $entity): array {
    // @todo This is called for every single entity insert, update or deletion
    //   on the whole site. Should maybe be cached?
    $datasource_id = 'workspaces_entity:' . $entity->getEntityTypeId();
    $entity_bundle = $entity->bundle();
    $has_bundles = $entity->getEntityType()->hasKey('bundle');

    /** @var \Drupal\search_api\IndexInterface[] $indexes */
    $indexes = [];
    try {
      $indexes = $this->entityTypeManager->getStorage('search_api_index')
        ->loadMultiple();
    }
      // @todo Replace with multi-catch once we depend on PHP 7.1+.
    catch (InvalidPluginDefinitionException $e) {
      // Can't really happen, but play it safe to appease static code analysis.
    }
    catch (PluginNotFoundException $e) {
      // Can't really happen, but play it safe to appease static code analysis.
    }

    foreach ($indexes as $index_id => $index) {
      // Filter out indexes that don't contain the datasource in question.
      if (!$index->isValidDatasource($datasource_id)) {
        unset($indexes[$index_id]);
      }
      elseif ($has_bundles) {
        // If the entity type supports bundles, we also have to filter out
        // indexes that exclude the entity's bundle.
        try {
          $config = $index->getDatasource($datasource_id)->getConfiguration();
        }
        catch (SearchApiException $e) {
          // Can't really happen, but play it safe to appease static code
          // analysis.
          unset($indexes[$index_id]);
          continue;
        }
        if (!Utility::matches($entity_bundle, $config['bundles'])) {
          unset($indexes[$index_id]);
        }
      }
    }

    return $indexes;
  }
}
